// src/Screen.hh
// (C) 2005 - 2017 Philip Endecott
// This file is part of Anyterm; see http://anyterm.org/

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef Screen_hh
#define Screen_hh

#include <vector>
#include <algorithm>
#include <memory>

#include "Cell.hh"


class Screen {
  int rows_;
  int cols_;
  int scrollback_;
  typedef std::vector<Cell> row_t;
  typedef std::vector<std::unique_ptr<row_t>> cells_t;
  cells_t cells;
  unsigned int wrap;

  // 'wrap' is used to implement scrolling without the need to move all the data.
  // The first row of the scrollback region is at cells[wrap].  Row n is at cells[wrap+n % scrollback+rows].
  // Private method normalise_wrap can be called to move things around so that wrap 
  // is zero, when necessary.  The external interface doesn't care about this.

public:
  Screen(int rows, int cols, int scrollback=0):
    rows_(rows), cols_(cols), scrollback_(scrollback),
    wrap(0),
    cursor_row(0), cursor_col(0), cursor_visible(true)
  {
    for (int r=0; r<scrollback+rows; ++r) {
      cells.push_back(std::make_unique<row_t>(cols));
    }
  }

  int rows() const {
    return rows_;
  }

  int cols() const {
    return cols_;
  }

  int scrollback() const {
    return scrollback_;
  }

  Cell& operator()(int r, int c)
  {
    return (*cells[row_idx(r)])[c];
  }

  const Cell& operator()(int r, int c) const
  {
    return (*cells[row_idx(r)])[c];
  }

  void scroll_down(int top, int bottom, int n=1)
  {
    // If we're asked to scroll the whole visible screen down, we scroll
    // into the scrollback region.  Otherwise, the scrollback region is
    // not changed.
    if (top==0 && bottom==rows()-1) {
      wrap = (wrap+n)%(rows()+scrollback());
    } else {
      normalise_wrap();
      std::rotate(cells.begin()+scrollback()+top, cells.begin()+scrollback()+top+n, 
                  cells.begin()+scrollback()+bottom+1);
    }
    for (int r=bottom+1-n; r<=bottom; ++r) {
      clear_row(r);
    }
  }

  void scroll_up(int top, int bottom, int n=1)
  {
    // Never touch the scrollback region.
    if (scrollback()==0 && top==0 && bottom==rows()-1) {
      wrap = (wrap-n)%(rows()+scrollback());
    } else {
      normalise_wrap();
      std::rotate(cells.begin()+scrollback()+top, cells.begin()+scrollback()+bottom+1-n,
                  cells.begin()+scrollback()+bottom+1);
    }
    for (int r=top; r<top+n; ++r) {
      clear_row(r);
    }
  }

  void resize(int new_rows, int new_cols)
  {
    if (new_rows != rows()) {
      normalise_wrap();
      if (new_rows < rows()) {
        cells.resize(scrollback()+new_rows);
      } else if (new_rows > rows()) {
        for (int r=rows(); r<new_rows; ++r) {
          cells.push_back(std::make_unique<row_t>(cols()));
        }
      }
      rows_ = new_rows;
      if (cursor_row >= new_rows) cursor_row = new_rows-1;
    }

    if (new_cols != cols()) {
      for (auto i = cells.begin(); i!= cells.end(); ++i) {
        auto& row = **i;
        if (new_cols < cols()) {
          row.resize(new_cols);
        } else if (new_cols > cols()) {
          for (int c=cols(); c<new_cols; ++c) {
            row.emplace_back();
          }
        }
      }
      cols_ = new_cols;
      if (cursor_col >= new_cols) cursor_col = new_cols-1;
    }
  }

  int cursor_row;
  int cursor_col;
  bool cursor_visible;


private:
  int row_idx(int r) const {
    return (r+scrollback()+wrap)%(rows()+scrollback());
  }

  void clear_row(int r) {
    row_t& row = *(cells[row_idx(r)]);
    for (int c=0; c<cols(); ++c) {
      row[c] = Cell();   // FIXME this should probably use the terminal's current attributes 
                         // (e.g. background colour)
    }
  }

  void normalise_wrap() {
    if (wrap==0) {
      return;
    }
    std::rotate(cells.begin(), cells.begin()+wrap, cells.end());
    wrap=0;
  }
};


#endif
