// src/SubProcess.cc
// (C) 2006 - 2017 Philip Endecott
// This file is part of Anyterm; see http://anyterm.org/

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "SubProcess.hh"

// For forkpty():
#if defined(__FreeBSD__)
#include <libutil.h>
#include <sys/ioctl.h>
#include <termios.h>
#elif defined(__OpenBSD__) || defined(__NetBSD__)
#include <termios.h>
#include <util.h>
#elif defined(__APPLE__)
#include <util.h>
#elif defined(__sun__)
#include "sun_forkpty.h"
#include <sys/tty.h>
#else
#include <pty.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>

#include "Exception.hh"

#include "set_thread_name.hh"

#include <sstream>


SubProcess_base::SubProcess_base(std::string command, int pty_rows, int pty_cols)
{
  struct winsize ws;
  ws.ws_row=pty_rows;
  ws.ws_col=pty_cols;
  ws.ws_xpixel=0;
  ws.ws_ypixel=0;

  pid = forkpty(&fd_num, NULL, NULL, &ws);
  if (pid==-1) {
    pbe::throw_ErrnoException("forkpty()");
  }
  if (pid==0) {
    setenv("TERM","linux",1);
    struct termios t;
    tcgetattr(0,&t);  // Could fail, but where would we send the error?
    t.c_cc[VERASE]=8; // Make ctrl-H (backspace) the erase character.
    tcsetattr(0,TCSANOW,&t); // ditto.
    execl("/bin/sh","/bin/sh","-c",command.c_str(),NULL);
    pbe::throw_ErrnoException("execl(/bin/sh -c "+command+")");  // pointless.
  }
}


SubProcess::SubProcess(onOutput_t onOutput_,
                       onError_t onError_,
		       std::string command,
		       int pty_rows, int pty_cols):
  SubProcess_base(command, pty_rows, pty_cols),
  fd(fd_num, false),
  terminating(false),
  onOutput(onOutput_),
  onError(onError_),
  output_processor(std::make_unique<std::thread>([this]{pbe::set_thread_name("output_proc");
                                                        process_output();}))
{}


SubProcess::~SubProcess()
{
  terminating = true;

  // We do two things to try to kill the subprocess: we close the fd,
  // which really ought to kill it, and we SIGHUP it.  The SIGHUP
  // by itself may not be sufficient if the process is nohup or
  // setuid or something.  The close by itself really should be
  // sufficient, but I'm keeping the SIGHUP because I'm paranoid.
  // The three results that we want are (a) the process dies,
  // (b) it does not become a zombie, and (c) the output processor
  // thread terminates so that we can join it.
  // There's a danger that we could close the fd and something else 
  // could open another fd with the same number, which the output 
  // processor could read.  I hope that's not a high probability.

  try {
    fd.close();
    // fd.close() can throw.
  } catch (...) {}
  kill(pid,SIGHUP);

  output_processor->join();
}


void SubProcess::resize(int rows, int cols)
{
  struct winsize ws;
  ws.ws_row=rows;
  ws.ws_col=cols;
  ws.ws_xpixel=0;
  ws.ws_ypixel=0;

  fd.ioctl(TIOCSWINSZ, &ws);

  // The kernel sends SIGWINCH.
}



void SubProcess::send(std::string s)
{
  fd.writeall(s);
}


void SubProcess::process_output(void)
{
  try { try {
    while(!terminating) {
      std::string s = fd.readsome();
      onOutput(s);
    }
  } RETHROW_MISC_EXCEPTIONS }
  catch (pbe::IOError&) {
    onError("Subprocess terminated");
    return;
  }
  catch (pbe::Exception& e) {
    std::ostringstream s;
    s << "Exception: ";
    e.report(s);
    onError(s.str());
    return;
  }
}

